/*
 *  Support file for the Philips 82c200 basic can controller
 *
 *  - currently tested for PORT At-mini-can boards
 *  
 *
 *
 *
 *
 *
 *
 *
 */

header "82c200conf" /{
  /*
   * default configuration
   *
   */

/*---------- Default Outc value 
 * this depends on the transceiver configuration
 * 
 * the port board uses optocoupler configuration as denoted in the philips
 * application notes, so the OUTC value is 0xfa
 *
 */

#define CAN_OUTC_VAL            0xfa



}/

/*======================================================================*/
file "82c200" /{

/* timing values */
uint8 CanTiming[9][2]={ {CAN_TIM0_10K,  CAN_TIM1_10K},
			{CAN_TIM0_20K,  CAN_TIM1_20K},
			{CAN_TIM0_40K,  CAN_TIM1_40K},
			{CAN_TIM0_50K,  CAN_TIM1_50K},
			{CAN_TIM0_125K, CAN_TIM1_125K},
			{CAN_TIM0_250K, CAN_TIM1_250K},
			{CAN_TIM0_500K, CAN_TIM1_500K},
			{CAN_TIM0_800K, CAN_TIM1_800K},
			{CAN_TIM0_1000K,CAN_TIM1_1000K}};

}/


/*======================================================================*/
header "82c200" /{
	
typedef unsigned char uint8;

extern uint8 CanTiming[9][2];

/*----------*/

typedef struct canregs {
	uint8    canctl;
	uint8    cancmd;
	uint8    canstat;
	uint8	 canirq;
	uint8	 canacc;
	uint8	 canmask;
	uint8	 cantim0;
	uint8	 cantim1;
	uint8	 canoutc;
	uint8	 cantest;
	uint8	 cantxdes1;
	uint8	 cantxdes2;
	uint8    cantxdata[8];
	uint8	 canrxdes1;
	uint8    canrxdes2;
	uint8    canrxdata[8];
	uint8    unused;
	uint8	 canclk; 	 
} canregs_t;

#define CAN_RANGE 0x20

/*--- Control Register ---------------------------------*/
 
#define CAN_TEST_MODE				(1<<7)
#define CAN_SPEED_MODE				(1<<6)
#define CAN_OVERRUN_INT_ENABLE			(1<<4)
#define CAN_ERROR_INT_ENABLE			(1<<3)
#define CAN_TRANSMIT_INT_ENABLE			(1<<2)
#define CAN_RECEIVE_INT_ENABLE			(1<<1)
#define CAN_RESET_REQUEST			(1<<0)

/*--- Command Register ------------------------------------*/
 
#define CAN_GOTO_SLEEP				(1<<4)
#define CAN_CLEAR_OVERRUN_STATUS		(1<<3)
#define CAN_RELEASE_RECEIVE_BUFFER		(1<<2)
#define CAN_ABORT_TRANSMISSION			(1<<1)
#define CAN_TRANSMISSION_REQUEST		(1<<0)

/*--- Status Register --------------------------------*/
 
#define CAN_BUS_STATUS 				(1<<7)
#define CAN_ERROR_STATUS			(1<<6)
#define CAN_TRANSMIT_STATUS			(1<<5)
#define CAN_RECEIVE_STATUS			(1<<4)
#define CAN_TRANSMISSION_COMPLETE_STATUS	(1<<3)
#define CAN_TRANSMIT_BUFFER_ACCESS		(1<<2)
#define CAN_DATA_OVERRUN			(1<<1)
#define CAN_RECEIVE_BUFFER_STATUS		(1<<0)

/*--- Status Register --------------------------------*/
 
#define CAN_BUS_STATUS_BIT 			(1<<7)
#define CAN_ERROR_STATUS_BIT			(1<<6)
#define CAN_TRANSMIT_STATUS_BIT			(1<<5)
#define CAN_RECEIVE_STATUS_BIT			(1<<4)
#define CAN_TRANSMISSION_COMPLETE_STATUS_BIT	(1<<3)
#define CAN_TRANSMIT_BUFFER_ACCESS_BIT		(1<<2)
#define CAN_DATA_OVERRUN_BIT			(1<<1)
#define CAN_RECEIVE_BUFFER_STATUS_BIT		(1<<0)

/*--- Interrupt Register -----------------------------------*/
 
#define CAN_WAKEUP_INT				(1<<4)
#define CAN_OVERRUN_INT				(1<<3)
#define CAN_ERROR_INT				(1<<2)
#define CAN_TRANSMIT_INT			(1<<1)
#define CAN_RECEIVE_INT 			(1<<0)

/*--- Output Control Register -----------------------------------------*/
/*
 *	7	6	5	4	3	2	1	0
 * 	OCTP1	OCTN1	OCPOL1	OCTP0	OCTN0	OCPOL0	OCMODE1	OCMODE0
 *	----------------------  ----------------------  ---------------
 *	    TX1 Output		    TX0 Output		  programmable
 *	  Driver Control	  Driver Control	  output functions
 *
 *	MODE
 *	OCMODE1	OCMODE0
 *	  0	  1	Normal Mode; TX0, TX1 bit sequenze TXData
 *	  1	  1	Normal Mode; TX0 bit sequenze, TX1 busclock TXCLK
 *	  0	  0	Biphase Mode
 *	  1	  0	Test Mode; TX0 bit sequenze, TX1 COMPOUT
 *
 *	In normal Mode Voltage Output Levels depend on 
 *	Driver Characteristic: OCTPx, OCTNx
 *	and programmed Output Polarity: OCPOLx
 *
 *	Driver Characteristic
 *	OCTPx	OCTNx
 *	  0	 0	always Floating Outputs,
 *	  0	 1	Pull Down
 *	  1	 0	Pull Up
 *	  1	 1	Push Pull
 */
 
/*--- Output control register --------------------------------*/

#define CAN_OCTP1			(1<<7)
#define CAN_OCTN1			(1<<6)
#define CAN_OCPOL1			(1<<5)
#define CAN_OCTP0			(1<<4)
#define CAN_OCTN0			(1<<3)
#define CAN_OCPOL0			(1<<2)
#define CAN_OCMODE1			(1<<1)
#define CAN_OCMODE0			(1<<0)

/*--- Remote Request ---------------------------------*/
/*    Notes:	RTR is Bit 5 in TXDES1.
 */
#define CAN_RTR					(1<<4)

#define CANout(bd,adr,v) outb( v, (int) &((canregs_t *)Base[bd])->adr )
#define CANin(bd,adr)    inb( (int) &((canregs_t *)Base[bd])->adr  )
#define CANset(bd,adr,m)   outb(inb((int) &((canregs_t *)Base[bd])->adr) | m ,(int) &((canregs_t *)Base[bd])->adr )
#define CANreset(bd,adr,m) outb(inb((int) &((canregs_t *)Base[bd])->adr) & ~m,(int) &((canregs_t *)Base[bd])->adr )
#define CANtest(bd,adr,m) (inb( (int) &((canregs_t *)Base[bd])->adr  ) & m )

/* the base address register array */
extern int Base[];
/*---------- Timing values */

/* the timings are valid for clock 8Mhz */
#define CAN_TIM0_10K		  49
#define CAN_TIM1_10K		0x1c
#define CAN_TIM0_20K		  24	
#define CAN_TIM1_20K		0x1c
#define CAN_TIM0_40K		0x89	/* Old Bit Timing Standard of port */
#define CAN_TIM1_40K		0xEB	/* Old Bit Timing Standard of port */
#define CAN_TIM0_50K		   9
#define CAN_TIM1_50K		0x1c
#define CAN_TIM0_125K		   3
#define CAN_TIM1_125K		0x1c
#define CAN_TIM0_250K		   1
#define CAN_TIM1_250K		0x1c
#define CAN_TIM0_500K		   0
#define CAN_TIM1_500K		0x1c
#define CAN_TIM0_800K		   0
#define CAN_TIM1_800K		0x16
#define CAN_TIM0_1000K		   0
#define CAN_TIM1_1000K		0x14

}/
/*======================================================================*/


file "82c200_irq" /{

void Can_c200_Interrupt ( int irq, void *unused, struct pt_regs *ptregs ) {
   int minor = irq2minormap[irq];
   register i;
   int irqsrc,dummy;
   msg_fifo_t *RxFifo = &Rx_Buf[minor]; 
   msg_fifo_t *TxFifo = &Tx_Buf[minor];
   msg_filter_t *RxPass = &Rx_Filter[minor];
   unsigned int id; 
   canmsg_t msg;
   
#ifdef IDEBUG   
   printk("CAN Interrupt %d (minor=%d Rxhead=%d Rxtail=%d  Txhead=%d Txtail=%d )!\n",
           irq,minor,RxFifo->head, RxFifo->tail,
           TxFifo->head, TxFifo->tail
           );
#endif

   irqsrc = CANin(minor,canirq);

   /*========== receive interrupt */
   if( irqsrc & CAN_RECEIVE_INT ) {

	dummy  = CANin(minor, canrxdes2 );
#if 1
	id = CANin(minor, canrxdes1 ) << 3 | (dummy & 0xe0) >> 5 ;
	if(RxPass->use) {
		if(!RxPass->filter[id].enable) {
#if IDEBUG
			printk("discarded %d\n",id); 
#endif
			goto DiscardRX;
		}
	} 
	(RxFifo->data[RxFifo->head]).id = id;
#else
        (RxFifo->data[RxFifo->head]).id = CANin(minor, canrxdes1 ) << 3 | (dummy & 0xe0) >> 5 ;
#endif

#if 0
printk("itx: 1=0x%x 2=0x%x id=%d\n", CANin(minor, canrxdes1),CANin(minor, canrxdes2),
CANin(minor, canrxdes1 ) << 3 | (dummy & 0xe0) >> 5 );
#endif
        dummy  &= 0x0f; /* strip length code */ 
        (RxFifo->data[RxFifo->head]).length = dummy;
        for( i=0;i<dummy && i<8 ;i++) 
	   (RxFifo->data[RxFifo->head]).data[i] = CANin(minor,canrxdata[i]);
	RxFifo->status = BUF_OK;
        RxFifo->head = ++(RxFifo->head) % MAX_BUFSIZE;
	if(RxFifo->head == RxFifo->tail) {
		printk("%d RX: FIFO overrun",minor);
		RxFifo->status = BUF_OVERRUN;
        } 

        /*---------- kick the select() call  -*/
        wake_up_interruptible(  &CanWait[minor] ); 
DiscardRX:
        CANout(minor, cancmd, CAN_RELEASE_RECEIVE_BUFFER | CAN_CLEAR_OVERRUN_STATUS );

   }



   /*========== transmit interrupt */
   if( irqsrc & CAN_TRANSMIT_INT ) {
	uint8 tx2reg;
	if( TxFifo->head == TxFifo->tail  ) {
	   TxFifo->status = BUF_EMPTY;
           TxFifo->active = 0;
           goto tx_done;
	}
#if 0
printk("txi: id=%d msg='%s' \n", (TxFifo->data[TxFifo->tail]).id, (TxFifo->data[TxFifo->tail]).data );
#endif
        CANout(minor, cantxdes1, (uint8)(( (TxFifo->data[TxFifo->tail]).id ) >> 3));
        tx2reg =  (( (TxFifo->data[TxFifo->tail]).id << 5 ) & 0xe0) 
                     | ((uint8) (TxFifo->data[TxFifo->tail]).length  ) ;
        if( (TxFifo->data[TxFifo->tail]).rtr ) 
	    tx2reg |= CAN_RTR;
        CANout(minor, cantxdes2, tx2reg );

        for( i=0; i < (TxFifo->data[TxFifo->tail]).length ; i++) 
	    CANout(minor, cantxdata[i], (TxFifo->data[TxFifo->tail]).data[i]);
        
        CANout(minor, cancmd, CAN_TRANSMISSION_REQUEST ); 
	TxFifo->tail = ++(TxFifo->tail) % MAX_BUFSIZE;
   }
tx_done:
   /*========== error status */
   if( irqsrc & CAN_ERROR_INT ) {
	printk("%d: Tx err!\n",minor);
        TxErr[minor]++;
	
   }
   if( irqsrc & CAN_OVERRUN_INT ) {
	printk("%d: overrun!\n",minor);
        Overrun[minor]++;
   }


}


}/


/*======================================================================*/

file "82c200funcs" /{

extern int Can_c200_ChipReset (int);
extern int Can_c200_SetTiming (int,int);
extern int Can_c200_StartChip(int);
extern int Can_c200_StopChip(int);
extern int Can_c200_SetMask(int,int);
extern int Can_c200_SetOMode(int,int);
extern int Can_c200_SendMessage(int, canmsg_t *, int );
extern int Can_c200_GetMessage(int , canmsg_t *);
extern void Can_c200_Interrupt( int irq, void *unused, struct pt_regs *ptregs );

can_atomic_func_t p82c200_ops = {
   Can_c200_ChipReset,
   Can_c200_SetTiming,
   Can_c200_StartChip,
   Can_c200_StopChip,
   Can_c200_SetMask,
   Can_c200_SetOMode,
   Can_c200_SendMessage,
   Can_c200_GetMessage,
   Can_c200_Interrupt
};


}/


/@
  Resetting 82C200 Controller
  board is the board id
@/
func c200_ChipReset (int board) /{

	uint8 status;
	int i;

	printk("Resetting Board %d on 0x%x \n", board, Base[board]);
	DBGprint(DBG_DATA,("cmdr=0x%x\n", &((canregs_t *)Base[board])->cancmd ));

	CANset(board, canctl, CAN_RESET_REQUEST );
	for(i=0;i<100;i++) SLOW_DOWN_IO;

	status = CANin(board, canstat);
	DBGprint(DBG_DATA,("status=0x%x ctl=0x%x",status,
		CANin(board, canctl) ));
	if( ! (CANin(board, canctl) & CAN_RESET_REQUEST ) ){
		printk("ERROR: no board present!");
		DBGout();return -1;
	}
	
        /* Board specific output control */
	DBGprint(DBG_DATA,("outc=0x%x",CAN_OUTC_VAL ));
	CANout(board, canoutc, CAN_OUTC_VAL );

	Can_c200_SetTiming(board, Baud[board]    );
	Can_c200_SetMask  (board, AccMask[board] );


	DBGout();
	return status;
}/


/@
  Set timing values
@/
func c200_SetTiming (int board,int baud) /{
int i;
    DBGprint(DBG_DATA,("baud[%d]=%d",board,baud));
    switch(baud)
    {
	case   10: i = 0; break;
	case   20: i = 1; break;
	case   40: i = 2; break;
	case   50: i = 3; break;
	case  125: i = 4; break;
	case  250: i = 5; break;
	case  500: i = 6; break;
	case  800: i = 7; break;
	case 1000: i = 8; break;
	default  : 
		DBGout();
		return(-EINVAL);
    }
    CANout(board,canclk,0);
    CANout(board,cantim0, (uint8) CanTiming[i][0]);
    CANout(board,cantim1, (uint8) CanTiming[i][1]);
    DBGprint(DBG_DATA,("tim0=0x%x tim1=0x%x",CANin( board,cantim0),CANin( board,cantim1)) );

}/
/*======================================================================*/
/@
 Start Controller
@/
func c200_StartChip (int board) /{
	int i;
	RxErr[board] = TxErr[board] = 0L;
/*
	CANout( board,cancmd, (CAN_RELEASE_RECEIVE_BUFFER 
                              | CAN_CLEAR_OVERRUN_STATUS) ); 
*/
	CANreset( board, canctl, CAN_RESET_REQUEST );
	for(i=0;i<100;i++) SLOW_DOWN_IO;
	CANin(board,canirq);

CANset( board, canctl, CAN_RECEIVE_INT_ENABLE );
	DBGprint(DBG_DATA,("start ctl=0x%x",CANin( board,canctl)) );
}/

/@ 
 Stop Controller 
 @/
func c200_StopChip (int board) /{
	CANset(board, canctl, CAN_RESET_REQUEST );
}/

/@ 
 Set ACCMASK and ACCR
 @/
func c200_SetMask (int board, int arg) /{
	DBGprint(DBG_DATA,("acc=0x%x mask=0x%x", 
			  (arg & 0xffff) >> 8,(arg & 0xff)  ));
	CANout(board, canacc, (arg & 0xffff) >> 8 );	
	CANout(board, canmask, (arg & 0xff)  );
}/

/@ 
 Set OUTC register
 @/
func c200_SetOMode (int board, int arg) /{
	DBGprint(DBG_DATA,("outc=0x%x",arg));
	CANout(board,canoutc, arg );
}/

/@

 Send CAN Message to Tx Buffer with polling

 @/
func c200_SendMessage(int board, canmsg_t *tx, int wait) /{
  int i=0,cycle=0,k=0;
  uint8 tx2reg,stat;


  if(wait) LDDK_START_TIMER(Timeout[board]);
  while ( ! (stat=CANin(board,canstat)) & CAN_TRANSMIT_BUFFER_ACCESS && ! LDDK_TIMEDOUT ) {
	if( need_resched ) schedule();
  }

  if( ! LDDK_TIMEDOUT  ) {
    DBGprint(DBG_DATA,("0x%x: tx.rtr=%d tx.id=0x%x tx.length=%d stat=0x%x",
                   Base[board],tx->rtr,tx->id,tx->length,stat));

    CANout(board, cantxdes1, (uint8)((tx->id) >> 3) );
    tx2reg =  ((tx->id << 5 ) & 0xe0) | ((uint8)(tx->length)) ;
    if( tx->rtr ) 
	    tx2reg |= CAN_RTR;
    CANout(board, cantxdes2, tx2reg );

    for( i=0; i < tx->length; i++) {
	CANout(board, cantxdata[i], tx->data[i]);
	DBGprint(DBG_DATA,("tx[%d]: 0x%x ('%c')",i,tx->data[i],tx->data[i]));
    }
    CANout(board, cancmd, CAN_TRANSMISSION_REQUEST );
    if( wait ) {
       while( ! (CANin(board, canstat) & CAN_TRANSMISSION_COMPLETE_STATUS) 
           && ! LDDK_TIMEDOUT ) cycle++;
       if( LDDK_TIMEDOUT ) {
       DBGprint(DBG_DATA,("Timeout! stat=0x%x",CANin(board, canstat  )));	
	/*CANout(board, cancmd, CAN_ABORT_TRANSMISSION );*/
    

       /*CANout( board,cancmd, (CAN_RELEASE_RECEIVE_BUFFER 
	                              | CAN_CLEAR_OVERRUN_STATUS) );*/
       }
    }
  }  else {
    printk("Tx: Timeout! stat=0x%x\n",stat);
    i= -1;
  }
  if(wait) LDDK_STOP_TIMER();
  DBGout();return i;
}/


/@
 Get Message from Receive Buffer with polling
@/
func c200_GetMessage(int board, canmsg_t *rx ) /{
  uint8 dummy,stat;
  int i=0;
  
  LDDK_START_TIMER(Timeout[board]);
  stat = CANin(board, canstat  );
  DBGprint(DBG_DATA,("0x%x:stat=0x%x ctl=0x%x",Base[board],stat,CANin(board,canctl)));
  rx->rtr=0;
  rx->length=0;

  if( stat & CAN_DATA_OVERRUN ) {
	DBGprint(DBG_DATA,("Rx: overrun!\n"));
	Overrun[board]++;
  }

  if( stat & CAN_RECEIVE_BUFFER_STATUS ) {
    dummy  = CANin(board, canrxdes2 );
    rx->id = CANin(board, canrxdes1 ) << 3 | (dummy & 0xe0) >> 5 ;
    dummy  &= 0x0f; /* strip length code */ 
    rx->length = dummy;
    DBGprint(DBG_DATA,("rx.id=0x%x rx.length=0x%x",rx->id,dummy));

    for( i=0;i<dummy && i<8 ;i++) {
	rx->data[i] = CANin(board,canrxdata[i]);
	DBGprint(DBG_DATA,("rx[%d]: 0x%x ('%c')",i,rx->data[i],rx->data[i] ));
    }
    CANout(board, cancmd, CAN_RELEASE_RECEIVE_BUFFER | CAN_CLEAR_OVERRUN_STATUS );
  } else {
    i=0;
  }
  LDDK_STOP_TIMER();
  DBGout();
  return i;
}/


